;;;; remote-mailbox-server.scm
;;;; Kon Lovett, Sep '09

;; Issues
;;
;; - Currently the server output port is ignored since the client input port is
;; closed!

(module remote-mailbox-server

  (;export
    ;; Common
    ; Parameters
    default-remote-mailbox-tcp-port
    default-remote-mailbox-hostname
    ;; Server
    ; Parameters
    default-remote-mailbox-deserializer
    default-remote-mailbox-listen
    default-remote-mailbox-auto-create?
    ; Operations
    make-remote-mailbox-server
    make-remote-mailbox-server-thread
    local-mailbox/server
    drop-local-mailbox!/server
    local-mailbox-names/server
    remote-mailbox-server-run!
    remote-mailbox-server-start!
    ;; Queries
    remote-mailbox-server?
    remote-mailbox-server-name
    remote-mailbox-server-auto-create?
    remote-mailbox-server-deserializer
    remote-mailbox-server-listener
    remote-mailbox-server-request-limit
    remote-mailbox-server-debug
    ;; Convenience
    local-mailbox-server
    local-mailbox-thread
    local-mailbox-start!
    local-mailbox)

  (import scheme
          chicken
          (only srfi-18 make-thread thread-start! mutex-name)
          (only data-structures identity)
          tcp
          (only miscmacros define-parameter while)
          tcp-server
          s11n
          mailbox
          lookup-table-synch
          type-checks
          type-errors
          conditions
          remote-mailbox-common)

  (require-library srfi-18 data-structures tcp
                   tcp-server s11n mailbox miscmacros lookup-table-synch type-checks conditions
                   remote-mailbox-common)

;;; Conditions

(define (make-remote-mailbox-server-condition msg rmbs args kind)
  (make-exn-condition+
    'remote-mailbox-tcp-server
    msg
    (cons (remote-mailbox-server-name rmbs) args)
    'remote-mailbox kind) )

(define (remote-mailbox-server-mailbox-exception rmbs . args)
  (abort (make-remote-mailbox-server-condition "no such mailbox" rmbs args 'mailbox)) )

(define (remote-mailbox-server-request-exception rmbs . args)
  (abort (make-remote-mailbox-server-condition "unknown remote mailbox client request"  rmbs args 'request)) )

;;; Server Side

;;

(define-record-type remote-mailbox-server
  (*make-remote-mailbox-server dctm nm autof desrl srvr lstnr rlim dbg thrd)
  remote-mailbox-server?
  (dctm remote-mailbox-server-dict/synch)
  (nm remote-mailbox-server-name)
  (autof remote-mailbox-server-auto-create?)
  (desrl remote-mailbox-server-deserializer)
  (srvr remote-mailbox-server-tcp-server remote-mailbox-server-tcp-server-set!)
  (lstnr remote-mailbox-server-listener)
  (rlim remote-mailbox-server-request-limit)
  (dbg remote-mailbox-server-debug) )

;; Parameters

(define-constant default-request-count-limit 10000)

(define-parameter default-remote-mailbox-deserializer #f
  (lambda (x)
  	(cond ((procedure? x)   x)
          ((not x)          #f)
          (else
            (warning-argument-type 'default-remote-mailbox-deserializer x 'procedure)
            (default-remote-mailbox-deserializer) ) ) ) )

(define-parameter default-remote-mailbox-listen tcp-listen
  (lambda (x)
  	(cond ((procedure? x) x)
          (else
            (warning-argument-type 'default-remote-mailbox-listen x 'procedure)
            (default-remote-mailbox-listen) ) ) ) )

(define-parameter default-remote-mailbox-auto-create? #t identity)

;; Support

(define-check+error-type remote-mailbox-server)

(define (*local-mailbox/server rmbs name create?)
	(dict-indempotent-ref!/synch
    (remote-mailbox-server-dict/synch rmbs)
    name
    (lambda (def) (if create? (make-mailbox name) def))) )

(define ((make-remote-mailbox-server-thunk rmbs))
  (while (not (eof-object? (peek-char (current-input-port))))
    (let ((req (deserialize (current-input-port) (remote-mailbox-server-deserializer rmbs))))
      (cond ((eq? (void) req) ) ;ignore void transmissions
            ((remote-mailbox-packet? req)
              (let* ((nam (remote-mailbox-packet-key req))
                     (lmb (*local-mailbox/server rmbs nam (remote-mailbox-server-auto-create? rmbs))) )
                (if lmb (mailbox-send! lmb (remote-mailbox-packet-value req))
                    (remote-mailbox-server-mailbox-exception rmbs nam)) ) )
            (else
              (remote-mailbox-server-request-exception rmbs req) ) ) ) ) )

(define (*remote-mailbox-server-run! rmbs)
	((remote-mailbox-server-tcp-server rmbs) (remote-mailbox-server-debug rmbs)) )

(define (*make-remote-mailbox-server-thread rmbs)
  (make-thread (lambda () (*remote-mailbox-server-run! rmbs)) (remote-mailbox-server-name rmbs)) )

(define (*remote-mailbox-server-start! rmbs)
	(thread-start! (*make-remote-mailbox-server-thread rmbs)) )

;; Exported

(define (make-remote-mailbox-server
          #!key (tcp-port (default-remote-mailbox-tcp-port))
                (listen (default-remote-mailbox-listen))
                (deserializer (default-remote-mailbox-deserializer))
                (name (gensym 'remote-mailbox-server:))
                (auto-create? (default-remote-mailbox-auto-create?))
                (request-limit default-request-count-limit)
                debug)
	(check-tcp-port 'make-remote-mailbox-server tcp-port 'tcp-port)
	(check-procedure 'make-remote-mailbox-server listen 'listen)
	(when deserializer (check-procedure 'make-remote-mailbox-server deserializer 'deserializer))
	(check-fixnum 'make-remote-mailbox-server request-limit 'request-limit)
  (let* ((listener (listen tcp-port))
         (rmbs (*make-remote-mailbox-server (make-dict/synch) name auto-create?
                                            deserializer #f listener request-limit
                                            debug #f))
         (tcps (make-tcp-server listener
                                (make-remote-mailbox-server-thunk rmbs)
                                request-limit)) )
    (remote-mailbox-server-tcp-server-set! rmbs tcps)
    rmbs ) )

(define (make-remote-mailbox-server-thread rmbs)
	(check-remote-mailbox-server 'make-remote-mailbox-server-thread rmbs)
  (*make-remote-mailbox-server-thread rmbs) )

;;

(define (remote-mailbox-server-run! rmbs)
	(check-remote-mailbox-server 'remote-mailbox-server-run! rmbs)
	(*remote-mailbox-server-run! rmbs) )

(define (remote-mailbox-server-start! rmbs)
	(check-remote-mailbox-server 'remote-mailbox-server-start! rmbs)
	(*remote-mailbox-server-start! rmbs) )

;;

(define (local-mailbox/server rmbs name #!optional (create? (remote-mailbox-server-auto-create? rmbs)))
	(check-remote-mailbox-server 'local-mailbox/server rmbs)
	(check-mailbox-name 'local-mailbox/server name)
	(*local-mailbox/server rmbs name create?) )

(define (drop-local-mailbox!/server rmbs name)
	(check-remote-mailbox-server 'drop-local-mailbox!/server rmbs)
	(check-mailbox-name 'drop-local-mailbox!/server name)
	(dict-delete!/synch (remote-mailbox-server-dict/synch rmbs) name) )

(define (local-mailbox-names/server rmbs)
	(check-remote-mailbox-server 'remote-mailbox-server-mailbox-names rmbs)
	(dict-keys/synch (remote-mailbox-server-dict/synch rmbs)) )

;;; Convenience

(define local-mailbox-start!)
(define local-mailbox-server)
(define local-mailbox-thread)
(define local-mailbox)
(let ((rmbs #f) (thrd #f))

  (set! local-mailbox-start!
    (lambda (#!optional debug)
      (unless rmbs
        (set! rmbs (make-remote-mailbox-server #:name 'default-remote-mailbox-server #:debug debug))
        (set! thrd (*remote-mailbox-server-start! rmbs)) ) ) )

  (set! local-mailbox-server (lambda () rmbs))

  (set! local-mailbox-thread (lambda () thrd))

  (set! local-mailbox
    (lambda (name #!optional debug)
      (check-mailbox-name 'local-mailbox name)
      (unless rmbs (local-mailbox-start! debug))
      (*local-mailbox/server rmbs name #t) ) ) )

) ;module remote-mailbox-server
